/*
 * TransFig: Facility for Translating Fig code
 * Copyright (c) 1991 by Micah Beck
 * Parts Copyright (c) 1985-1988 by Supoj Sutanthavibul
 * Parts Copyright (c) 1989-2002 by Brian V. Smith
 * Parts Copyright (c) 2005-2006 by Eric Colin de Verdiere
 *
 * Any party obtaining a copy of these files is granted, free of charge, a
 * full and unrestricted irrevocable, world-wide, paid up, royalty-free,
 * nonexclusive right and license to deal in this software and
 * documentation files (the "Software"), including without limitation the
 * rights to use, copy, modify, merge, publish and/or distribute copies of
 * the Software, and to permit persons who receive copies from any such 
 * party to do so, with the only requirement being that this copyright 
 * notice remain intact.
 *
 */

/*      genfrg.c : frg support for fig2dev
 *
 *      Author: Eric Colin de Verdiere
 *      Version 1 (updated for version 1.1 of the xfigfrag package)
 * 	It uses the LaTeX and PostScript drivers to generate 
 *      LaTeX processed text for a Postscript figure.
 */

#include "fig2dev.h"
#include "genps.h"
#include "genpdf.h"
#include "object.h"
#include "texfonts.h"

#define TOP		840

#define	SWAP(x,y)	{tmp=x; x=y; y=tmp;}
#define TRANS(x,y)		(*translate_coordinates)(&x,&y)
#define TRANS2(x1,y1,x2,y2)	(*translate_coordinates)(&x1,&y1); \
				(*translate_coordinates)(&x2,&y2)
#define TRANSD(x,y)		(*translate_coordinates_d)(&x,&y)
#define	ABS(x)		(((x) >= 0)? (x): -(x))
#define round4(x)	((round(10000.0*(x))/10000.0))
#define round6(x)	((round(1000000.0*(x))/1000000.0))

extern double rad2deg;

extern void
     	geneps_option (),
	genps_start (),
        genlatex_option (),
        unpsfont ();

extern int
	genps_end ();

extern double		dash_mag;

extern char *ISO1toTeX[];
extern char *ISO2toTeX[];

extern Boolean	FontSizeOnly;	/* defined in setfigfont.c */

extern int (*translate_coordinates)();
extern int (*translate_coordinates_d)();

static	int	verbose = 0;

double		unitlength;
static  int 	encoding = 1;

static
translate2(xp, yp)
  int	*xp, *yp;
{
	*xp = *xp + 1;
	*yp = (double)(TOP - *yp -1);
	}

static
translate2_d(xp, yp)
  double	*xp, *yp;
{
	*xp = *xp + 1.0;
	*yp = (double)TOP - *yp -1.0;
	}

static int optarg_frg = 0;

void
genfrg_option(opt, optarg)
  char opt, *optarg;
{
  /* Have to choose a character for the option here.  Already used:
     fig2dev: DFGhjLmsV
     latex_option: adEFfLlmsv
     ps_eps_option:    ABbCceFfGgLlmMnNOpRSsTxyz
     union: AaBbCcDdEeFfGghjLlMmNOpRSsTVvxyz
     Options in fig2dev.c: 
     AaB:b:C:cD:d:E:eFf:G:g:hI:i:jkl:L:Mm:Nn:OoPp:q:R:rS:s:Tt:UVvX:x:Y:y:Wwz:?
     
     Remaining:
     with optarg: I, i, q, t, X, Y
     without optarg: k, o, P, r, U, W, w

     -I, -i, -X, -Y (used by MetaPost)
     -k (PCL)
     -o metapost
     -P, -W -w EPIC
     -r CGM
     -U not used

     Hence choose -i<number> where <number> indicates which text is not put
     in the frg file:
     0: nothing (default)
     1: text with special attribute
     2: text with non-special attribute
     3: text with PostScript font
  */
  if (opt == 'i')
    sscanf(optarg, "%d", &optarg_frg);
  else
    genlatex_option(opt, optarg);
}

void
genepsfrg_option(opt, optarg)
  char opt, *optarg;
{
  if (opt == 'i')
    sscanf(optarg, "%d", &optarg_frg);
  else
    geneps_option(opt, optarg);
}

void
genpdffrg_option(opt, optarg)
  char opt, *optarg;
{
  if (opt == 'i')
    sscanf(optarg, "%d", &optarg_frg);
  else
    genpdf_option(opt, optarg);
}

/* decides if the text has to be put in the .frg file or not, depending on
   its attributes and on the value of the option -i */
int
genfrg_frg_text(t)
  F_text	*t;
{
        switch (optarg_frg) {
          
        case 0:
          return(1);
        case 1:
          return(!(special_text(t)));
        case 2:
          return(special_text(t));
        case 3:
          return(!(psfont_text(t)));
        }
        return(1);
}

void
genfrg_start(objects)
  F_compound	*objects;
{
	int tmp;

	texfontsizes[0] = texfontsizes[1] = 
		TEXFONTSIZE(font_size != 0.0? font_size : DEFAULT_FONT_SIZE);

 	unitlength = mag/ppi;
	dash_mag /= unitlength*80.0;

	translate_coordinates = translate2;
	translate_coordinates_d = translate2_d;

	TRANS2(llx, lly, urx, ury);
	if (llx > urx) SWAP(llx, urx)
	if (lly > ury) SWAP(lly, ury)

	/* LaTeX start */
        fprintf(tfp, "\\@xfigfrag@version{1}%\n%\n");

	/* print any whole-figure comments prefixed with "%" */
	if (objects->comments) {
	    fprintf(tfp,"%%\n");
	    print_comments("% ",objects->comments, "");
	    fprintf(tfp,"%%\n");
	}

	fprintf(tfp, "\\@xfigfrag@init{%lisp}{%d}{%d}{%d}{%d}%%\n",
                (long) (round(4736286.72*unitlength)),
                urx-llx, ury-lly, llx, lly);
}

int
genfrg_end()
{
  fprintf(tfp, "\\@xfigfrag@end%%\n");
        /* all ok */
        return 0;
}

void genpsfrg_text(t)
F_text	*t;
{
  if (!genfrg_frg_text(t)) {
    genps_text(t);
  }
}

void
genfrg_text(t)
  F_text	*t;
{
  /* output according to the following format:
     \@xfigfrag@put{x}{y}{pos}{rot}{rigid?}{special?}{psfont?}{hidden?}%
     {depth}{height}{length}{not nfss?}%
     {fontsizeonly?}{texsize}{baselineskip}{texfamily}{texseries}%
     {texshape}{texfont}%
     {color}{text}
   */
	int   	x, y;
	char	*tpos;
	unsigned char	*cp;

        if (genfrg_frg_text(t)) {
          unpsfont(t);

          if (verbose) 
	    fprintf(tfp, "%%\n%% Fig TEXT object\n%%\n");
          
          /* print any comments prefixed with "%" */
          print_comments("% ",t->comments, "");
          
          x = t->base_x;
          y = t->base_y;
          TRANS(x, y);
          
          switch (t->type) {
            
          case T_LEFT_JUSTIFIED:
          case DEFAULT:
            tpos = "0";
            break;
            
          case T_CENTER_JUSTIFIED:
            tpos = "1";
            break;
            
          case T_RIGHT_JUSTIFIED:
            tpos = "2";
            break;
            
          default:
            fprintf(stderr, "Text incorrectly positioned\n");
            tpos = "0";	/* make left in this case */
          }
          
          /* passing all attributes of the text, even if not used currently. */
          fprintf(tfp, "\\@xfigfrag@put{%3d}{%3d}{%s}{%1f}{%3d}\%\n",
                  x,y, tpos, t->angle*180/M_PI);
          fprintf(tfp, "{%1d}{%1d}{%1d}", rigid_text(t), special_text(t),
                  psfont_text(t), hidden_text(t));
          fprintf(tfp, "{%d}{%1f}{%1f}\%\n", t->depth, t->height, t->length);
#ifdef NFSS
          fprintf(tfp, "{0}");
#else
          fprintf(tfp, "{1}");
#endif
          
          /* font and size */
          int texsize;
          double baselineskip;
          texsize = TEXFONTMAG(t);
          baselineskip = (texsize * 1.2);
          fprintf(tfp, "{%1d}{%d}{%.1f}{%s}{%s}{%s}{%s}", ( FontSizeOnly ),
                  texsize, baselineskip,
                  TEXFAMILY(t->font),TEXSERIES(t->font),
                  TEXSHAPE(t->font), TEXFONT(t->font));
          
          /* color */
          static char *colors[] = {
            "0,0,0",    /* black */
            "0,0,1",    /* blue */
            "0,1,0",    /* green */
            "0,1,1",    /* cyan */
            "1,0,0",    /* red */
            "1,0,1",    /* magenta */
            "1,1,0",    /* yellow */
            "1,1,1",    /* white */
            "0,0,.56",   /* dk blue */
            "0,0,.69",   /* md blue */
            "0,0,.82",   /* lt blue */
            "0,.56,0",   /* dk green */
            "0,.69,0",   /* md green */
            "0,.82,0",   /* lt green */
            "0,.56,.56", /* dk cyan */
            "0,.69,.69", /* md cyan */
            "0,.82,.82", /* lt cyan */
            ".56,0,0",   /* dk red */
            ".69,0,0",   /* md red */
            ".82,0,0",   /* lt red */
            ".56,0,.56", /* dk magenta */
            ".69,0,.69", /* md magenta */
            ".82,0,.82", /* lt magenta */
            ".5,.17,0",  /* dk brown */
            ".63,.25,0", /* md brown1 */
            ".75,.38,0", /* md brown1 */
            ".88,.44,0", /* lt brown */
            "1,.5,.5",   /* dk pink */
            "1,.63,.63", /* md pink1 */
            "1,.75,.75", /* md pink2 */
            "1,.88,.88", /* lt pink */
            "1,.84,0",   /* gold */
          };
          /* switch to (other) color */
          if (t->color < NUM_STD_COLS)
            fprintf(tfp, "{%s}",colors[t->color]);
          else
            fprintf(tfp, "{%.3f,%.3f,%.3f}",
                    user_colors[t->color-NUM_STD_COLS].r/256.0,
                    user_colors[t->color-NUM_STD_COLS].g/256.0,
                    user_colors[t->color-NUM_STD_COLS].b/256.0);
          fprintf(tfp, "\%\n");
          
          /* Text */
          fprintf(tfp, "{");
          if (!special_text(t))
            /* this loop escapes characters "%{}" */
            /* and deletes characters "~^\" */
            for(cp = (unsigned char*)t->cstring; *cp; cp++) {
              if (strchr("%{}", *cp)) 
                fputc('\\', tfp);
              if (strchr("~\\", *cp))
                fprintf(stderr,
                        "Bad character in text object '%c'\n" ,*cp);
               else
                fputc(*cp, tfp);
            }
          else 
            for(cp = (unsigned char*)t->cstring; *cp; cp++) {
#ifdef I18N
              extern Boolean support_i18n;
              if (support_i18n && (t->font <= 2))
                fputc(*cp, tfp);
              else
#endif
                if (*cp >= 0xa0) {
                  switch (encoding) {
                  case 0: /* no escaping */
                    fputc(*cp, tfp);
                    break;
                  case 1: /* iso-8859-1 */
                    fprintf(tfp, "%s", ISO1toTeX[(int)*cp-0xa0]);
                    break;
                  case 2: /* iso-8859-2 */
                    fprintf(tfp, "%s", ISO2toTeX[(int)*cp-0xa0]);
                    break;
                  }
                } else
                  fputc(*cp, tfp);
            }
          
          fprintf(tfp, "}\%\n\%\n");
	}
}

void genpsfrg_option(opt, optarg)
char opt, *optarg;
{
  genlatex_option(opt, optarg);
}

struct driver dev_frg = {
  	genfrg_option,
	genfrg_start,
	gendev_null,
	gendev_null,
	gendev_null,
	gendev_null,
	gendev_null,
	genfrg_text,
	genfrg_end,
	INCLUDE_TEXT
};

struct driver dev_epsfrg = {
  	genepsfrg_option, 	/* use eps so always exported in Portrait mode */
	genps_start,
	genps_grid,
	genps_arc,
	genps_ellipse,
	genps_line,
	genps_spline,
	genpsfrg_text,
	genps_end,
	INCLUDE_TEXT
};

struct driver dev_pdffrg = {
  	genpdffrg_option,
	genpdf_start,
	genps_grid,
	genps_arc,
	genps_ellipse,
	genps_line,
	genps_spline,
	genpsfrg_text,
	genpdf_end,
	INCLUDE_TEXT
};
